//============================================================================
// Name        : ehSpeechEnhancement.cpp
// Author      : Eui-Hyun Kim (uihyun@gmail.com)
// Version     : 2.12.4
// Copyright   : Okuno Laboratory, Kyoto University
// Description :
//============================================================================

#include <iostream>
#include <string>

#include <math.h> // besell
#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>
#include <libharkio/libharkio.h>

#include "fft.hpp"

using namespace std;
using namespace FD;

class SpeechEnhancement;

DECLARE_NODE(SpeechEnhancement);
/*Node
 *
 * @name SpeechEnhancement
 * @category HARK:BINAURAL+
 * @description Speech Enhancement
 *
 * @input_name NOISY_SPEECH_SPECTRUM
 * @input_type Matrix<complex<float> >
 * @input_description Windowed spectrum data. A row index is channel, and a column index is frequency.
 *
 * @output_name SPEECH_EHANCED_SPECTRUM
 * @output_type Matrix<complex<float> >
 * @output_description Windowed and speech-enhanced spectrum data . A row index is channel, and a column index is frequency.
 *
 * @parameter_name NOISE_REDUCTION_METHOD
 * @parameter_type string
 * @parameter_value MINIMUM_MEAN_SQUARE_ERROR
 * @parameter_list SPECTRAL_SUBTRACTION:WIENER_FILTERING:TWO_STEP_WIENER_FILTERING:MINIMUM_MEAN_SQUARE_ERROR
 * @parameter_description Noise reduction methods [default: MINIMUM_MEAN_SQUARE_ERROR].
 *
 * @parameter_name HARMONIC_REGENERATION
 * @parameter_type string
 * @parameter_value USE
 * @parameter_list USE:NONE
 * @parameter_description Harmonic regeneration after noise reduction [default: USE].
 *
 * @parameter_name NOISE_PERIOD
 * @parameter_type float
 * @parameter_value 1.0
 * @parameter_description Time to be regarded as "noise" from the first frame. [default: 1.0 seconds]
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The length in sample between a frame and a previous frame. [default: 160]
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz)[default: 16000].
 *
 END*/


class ModifiedBessel {

public:
  ModifiedBessel() {};
  ~ModifiedBessel() {};

  /* PURPOSE: Evaluate modified Bessel function In(x) and n=0.  */
  double besseli0(double x)
  {
    double ax;
    double ans;
    double y;

    if ((ax=fabs(x)) < 3.75) {
      y = x / 3.75;
      y = y * y;
      ans = 1.0+y*(3.5156229+y*(3.0899424+y*(1.2067492+y*(0.2659732+y*(0.360768e-1+y*0.45813e-2)))));
    } else {
      y = 3.75 / ax;
      ans = (exp(ax)/sqrt(ax))*(0.39894228+y*(0.1328592e-1+y*(0.225319e-2+y*(-0.157565e-2+y*(0.916281e-2+y*(-0.2057706e-1+y*(0.2635537e-1+y*(-0.1647633e-1+y*0.392377e-2))))))));
    }
    return ans;
  }

  /* PURPOSE: Evaluate modified Bessel function In(x) and n=1.  */
  double besseli1(double x)
  {
    double ax;
    double ans;
    double y;

    if ((ax=fabs(x)) < 3.75) {
      y = x / 3.75;
      y = y * y;
      ans = ax*(0.5+y*(0.87890594+y*(0.51498869+y*(0.15084934+y*(0.2658733e-1+y*(0.301532e-2+y*0.32411e-3))))));
    } else {
      y = 3.75 / ax;
      ans = 0.2282967e-1+y*(-0.2895312e-1+y*(0.1787654e-1-y*0.420059e-2));
      ans = 0.39894228+y*(-0.3988024e-1+y*(-0.362018e-2+y*(0.163801e-2+y*(-0.1031555e-1+y*ans))));
      ans *= (exp(ax)/sqrt(ax));
    }
    return x < 0.0 ? -ans : ans;
  }

  /* PURPOSE: Evaluate modified Bessel function In(x) for n >= 0*/
  double besseli(int n, double x)
  {
    int j;
    double bi,bim,bip,tox,ans;

    if (n == 0)
      return( besseli0(x) );
    if (n == 1)
      return( besseli1(x) );
  }
};


class SpeechEnhancement : public BufferedNode {

  // input
  int spectrumInputID;

  // output
  int enhancedSpectrumOutputID;

  // parameters
  String noiseReductionMethod;
  String useOfHarmonicRegeneration;
  int numberOfNoiseFrames;
  int advance;
  int samplingRate;


  // variables
  float** noisePowerSpectrum;

  // for two-step noise reduction
  float* previousTSNRpowerSpectrum;

  // for harmonic regeneration
  rfft<float> rfft1;
  float* tempFilterGain;
  float* tempEnhancedMagnitude;

public:

  SpeechEnhancement(string nodeName, ParameterSet params) : BufferedNode(nodeName, params)
  {
    spectrumInputID = addInput("NOISY_SPEECH_SPECTRUM");
    enhancedSpectrumOutputID = addOutput("SPEECH_EHANCED_SPECTRUM");
    noiseReductionMethod = object_cast<String>(parameters.get("NOISE_REDUCTION_METHOD"));
    useOfHarmonicRegeneration = object_cast<String>(parameters.get("HARMONIC_REGENERATION"));
    float noiseTime = dereference_cast<float>(parameters.get("NOISE_PERIOD"));
    advance = dereference_cast<int>(parameters.get("ADVANCE"));
    samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));

    numberOfNoiseFrames = int(noiseTime/float(advance)*float(samplingRate));
    if (numberOfNoiseFrames <= 0)	numberOfNoiseFrames = 1;

    noisePowerSpectrum = NULL;

    previousTSNRpowerSpectrum = NULL;

    tempFilterGain = NULL;
    tempEnhancedMagnitude = NULL;

    inOrder = true;
  }

  virtual ~SpeechEnhancement()
  {
    delete [] noisePowerSpectrum; // need programming one line for "delete [] noisePowerSpectrum[mic]"
    delete [] previousTSNRpowerSpectrum;
    delete [] tempFilterGain;
    delete [] tempEnhancedMagnitude;
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    /////////////////////////////////////////////////
    // Set input data
    RCPtr<Matrix<complex<float> > > spectrumInputPtr = getInput(spectrumInputID, count);
    Matrix<complex<float> >& spectrumInput = *spectrumInputPtr;
    int nLows = spectrumInput.nrows();
    int nCols = spectrumInput.ncols()-1; // -1 is necessary.

    /////////////////////////////////////////////////
    // Set output data
    RCPtr<Matrix<complex<float> > > enhancedSpectrumPtr(new Matrix<complex<float> >(nLows, nCols+1));
    (*outputs[enhancedSpectrumOutputID].buffer)[count] = enhancedSpectrumPtr;
    Matrix<complex<float> >& enhancedSpectrum = *enhancedSpectrumPtr;

    ////////////////////////////////////////////////////
    // Initialization
    if (count == 0)
      {
	noisePowerSpectrum = new float*[nLows];
	for (int mic = 0; mic < nLows; mic++)
	  {
	    noisePowerSpectrum[mic] = new float[nCols];
	    for (int sample = 0; sample < nCols; sample++)
	      noisePowerSpectrum[mic][sample] = 0.f;
	  }

	previousTSNRpowerSpectrum = new float[nCols];
	tempFilterGain = new float[nCols];
	tempEnhancedMagnitude = new float[nCols];

	for (int sample = 0; sample < nCols; sample++)
	  {
	    previousTSNRpowerSpectrum[sample] = 0.f;
	    tempFilterGain[sample] = 0.f;
	    tempEnhancedMagnitude[sample] = 0.f;
	  }
      }

    // for each spectrum-channel
    for (int mic = 0; mic < nLows; mic++)
      {
	////////////////////////////////////////////////
	// Noise spectrum extraction
	if (count < numberOfNoiseFrames)
	  {
	    // store input spectrum as output data
	    //(*outputs[enhancedSpectrumOutputID].buffer)[count] = spectrumInputPtr;

	    for (int sample = 0; sample < nCols; sample++)
	      {
		noisePowerSpectrum[mic][sample] += (spectrumInput(mic, sample).real() * spectrumInput(mic, sample).real() + spectrumInput(mic, sample).imag() * spectrumInput(mic, sample).imag()) / numberOfNoiseFrames;
		//					// store zero spectrum as output data
		//					enhancedSpectrum(mic, sample).real(0.f);
		//					enhancedSpectrum(mic, sample).imag(0.f);
	      }
	    continue; // go to "for (int mic = 0; mic < nLows; mic++)"
	  }
	////////////////////////////////////////////////

	for (int sample = 0; sample < nCols; sample++)
	  {
	    // extract powerSpectrum and phase information from noisy input data : pow(abs(spectrumInput), 2)
	    float noisyPowerSpectrum = spectrumInput(mic, sample).real() * spectrumInput(mic, sample).real() + spectrumInput(mic, sample).imag() * spectrumInput(mic, sample).imag();
	    float noisyPhase = atan2(spectrumInput(mic, sample).imag(), spectrumInput(mic, sample).real());

	    float enhancedMagnitude = 0.f;
	    float filterGain = 0.f;

	    // estimation of signal-to-noise (SNR) and the Wiener gain
	    float postSNR = noisyPowerSpectrum / noisePowerSpectrum[mic][sample];
	    float instantaneousSNR = postSNR - 1;
	    if (instantaneousSNR < 0) instantaneousSNR = 0.1e-38; // limitation to prevent distortion

	    if (noiseReductionMethod == "SPECTRAL_SUBTRACTION")
	      {
		float spectralSubtraction = noisyPowerSpectrum - noisePowerSpectrum[mic][sample];
		if (spectralSubtraction < 0) spectralSubtraction = 0.1e-38; // limitation to prevent distortion

		filterGain = instantaneousSNR;
		enhancedMagnitude = sqrt(spectralSubtraction);
	      }

	    if (noiseReductionMethod == "WIENER_FILTERING")
	      {
		// Decision Directed with Instantaneous SNR
		const float beta = 0.98f; // typically 0.98
		float prioDD_SNR = beta * ( previousTSNRpowerSpectrum[sample] / noisePowerSpectrum[mic][sample] ) + (1-beta) * instantaneousSNR;
		if (prioDD_SNR < 0) prioDD_SNR = 0.1e-38; // limitation to prevent distortion

		float WienerGain = prioDD_SNR / (1+prioDD_SNR);
		if (WienerGain < 0) WienerGain = 0.1e-38; // limitation to prevent distortion

		filterGain = WienerGain;
		enhancedMagnitude = WienerGain * sqrt(noisyPowerSpectrum);
		previousTSNRpowerSpectrum[sample] = pow(enhancedMagnitude, 2);
	      }

	    if (noiseReductionMethod == "TWO_STEP_WIENER_FILTERING")
	      {
		// Decision Directed with Instantaneous SNR
		const float beta = 0.98f; // typically 0.98
		float prioDD_SNR = beta * ( previousTSNRpowerSpectrum[sample] / noisePowerSpectrum[mic][sample] ) + (1-beta) * instantaneousSNR;
		if (prioDD_SNR < 0) prioDD_SNR = 0.1e-38; // limitation to prevent distortion

		float firstWienerGain = prioDD_SNR / (1+prioDD_SNR);
		if (firstWienerGain < 0) firstWienerGain = 0.1e-38; // limitation to prevent distortion

		float prioTSNR_SNR = pow(firstWienerGain * sqrt(noisyPowerSpectrum), 2) / noisePowerSpectrum[mic][sample];
		if (prioTSNR_SNR < 0) prioTSNR_SNR = 0.1e-38; // limitation to prevent distortion

		float secondWienerGain = prioTSNR_SNR / (1+prioTSNR_SNR);
		if (secondWienerGain < 0) secondWienerGain = 0.1e-38; // limitation to prevent distortion

		filterGain = secondWienerGain;
		enhancedMagnitude = secondWienerGain * sqrt(noisyPowerSpectrum);
		previousTSNRpowerSpectrum[sample] = pow(enhancedMagnitude, 2);
	      }

	    if (noiseReductionMethod == "MINIMUM_MEAN_SQUARE_ERROR")
	      {
#if 1
		// Decision Directed with Instantaneous SNR
		const float beta = 0.98f; // typically 0.98
		float prioDD_SNR = beta * ( previousTSNRpowerSpectrum[sample] / noisePowerSpectrum[mic][sample] ) + (1-beta) * instantaneousSNR;
		if (prioDD_SNR < 0) prioDD_SNR = 0.1e-38; // limitation to prevent distortion

		ModifiedBessel bessel;
		float Vk = (prioDD_SNR/(1+prioDD_SNR))*instantaneousSNR; // SNRprio./(1+SNRprio).*SNRpost
		const float gamma1p5 = 0.8862;
		float MMSE_Gain = (gamma1p5*sqrt(Vk)/instantaneousSNR)*exp(-Vk/2)*((1+Vk)*bessel.besseli(0,Vk/2)+Vk*bessel.besseli(1,Vk/2));
		if (isnan(MMSE_Gain)||isinf(MMSE_Gain)) // Sometimes (occasionally) Gain values become infinite which are replaced by WienerGain in that case, that is:
		  MMSE_Gain = prioDD_SNR / (1+prioDD_SNR); // Wiener Filter for out of range values
#else
		const float lambda = 3.f;
		const float beta1 = 0.5f;
		const float beta2 = 1.f;
		float MMSE_Gain = 1 - lambda * pow( pow(1/(1+instantaneousSNR), beta1), beta2);
#endif
		if (MMSE_Gain < 0) MMSE_Gain = 0.1e-38; // limitation to prevent distortion
		filterGain = MMSE_Gain;
		enhancedMagnitude = MMSE_Gain * sqrt(noisyPowerSpectrum);
	      }

	    if (useOfHarmonicRegeneration == "USE")
	      {
		tempFilterGain[sample] = filterGain;
		tempEnhancedMagnitude[sample] = enhancedMagnitude;
	      }

	    // store enhanced-speech spectrum with its original phase information as output data
	    // enhancedMagnitude * exp(1i*noisyPhase) = enhancedMagnitude * ( cos(noisyPhase) + 1i*sin(noisyPhase) )
	    enhancedSpectrum(mic, sample).real(enhancedMagnitude * cos(noisyPhase));
	    enhancedSpectrum(mic, sample).imag(enhancedMagnitude * sin(noisyPhase));

	  } // for (int sample = 0; sample < nCols; sample++)

	//if (useOfHarmonicRegeneration == "USE" && noiseReductionMethod != "SPECTRAL_SUBTRACTION")
	if (useOfHarmonicRegeneration == "USE")
	  {
	    // ifft and fft for appying half wave rectificaiton
	    rfft1.initialize(nCols*2);

	    for (int sample = 0; sample < nCols; sample++)
	      rfft1.A[sample] = enhancedSpectrum(mic, sample);
	    rfft1.synthesize();

	    for (int sample = 0; sample < nCols*2; sample++)
	      if (rfft1.a[sample] < 0) // half wave rectification
		rfft1.a[sample] = 0.f;
	      else
		rfft1.a[sample] = rfft1.a[sample] * (1.0f / rfft1.tlength());
	    rfft1.analyze();

	    // processing harmonic regeneration
	    for (int sample = 0; sample < nCols; sample++)
	      {
		// extract powerSpectrum and phase information from noisy input data
		float noisyPowerSpectrum = spectrumInput(mic, sample).real() * spectrumInput(mic, sample).real() + spectrumInput(mic, sample).imag() * spectrumInput(mic, sample).imag();
		float noisyPhase = atan2(spectrumInput(mic, sample).imag(), spectrumInput(mic, sample).real());

		float harmonicRegeneratedMagnitude = 0.f;

		// extract powerSpectrum and phase information from noisy input data
		float tempEnhancedPowerSpectrum = pow(tempEnhancedMagnitude[sample], 2);
		float halfWavePowerSpectrum = rfft1.A[sample].real() * rfft1.A[sample].real() + rfft1.A[sample].imag() * rfft1.A[sample].imag();

		float prioHRNR_SNR = (tempFilterGain[sample] * tempEnhancedPowerSpectrum + (1-tempFilterGain[sample]) * halfWavePowerSpectrum) / noisePowerSpectrum[mic][sample];
		if (prioHRNR_SNR < 0) prioHRNR_SNR = 0.1e-38; // limitation to prevent distortion

		float HRNR_Gain = prioHRNR_SNR / (1+prioHRNR_SNR); // Wiener filtering
		if (HRNR_Gain < 0) HRNR_Gain = 0.1e-38; // limitation to prevent distortion

		harmonicRegeneratedMagnitude = HRNR_Gain * sqrt(noisyPowerSpectrum);

		// store enhanced-speech spectrum with its original phase information after harmonic regeneration as output data
		enhancedSpectrum(mic, sample).real(harmonicRegeneratedMagnitude * cos(noisyPhase));
		enhancedSpectrum(mic, sample).imag(harmonicRegeneratedMagnitude * sin(noisyPhase));
	      }
	  } // if (useOfHarmonicRegeneration == "USE" && noiseReductionMethod != "SPECTRAL_SUBTRACTION")
      } // for (int mic = 0; mic < nLows; mic++)

  }
};









